﻿//
// Немного допиливаем ядро яваскрипта и RightJS
//
RightJS.Element.include({
    text: function(content) {
        if(Browser.IE) {
            return (typeof(content) === 'undefined') ? this._.innerText : (this._.innerText = content);
        } else {
            return (typeof(content) === 'undefined') ? this._.textContent : (this._.textContent = content);
        }
    }
});
String.include({
    // from Prototype.js
    escapeHTML: function() {
        var div = document.createElement('div');
        var text = document.createTextNode(this);
        div.appendChild(text);
        return div.innerHTML;
    },

    // from Prototype.js
    unescapeHTML: function() {
        var div = document.createElement('div');
        div.innerHTML = this.stripTags();
        return div.childNodes[0].nodeValue;
    },

    toHTML: function() {
        return this.escapeHTML().replace(/\ /g, '&nbsp;').replace(/\n/g, '<br/>');
    },

    multiply: function(n) {
        var res = "";
        for(var i = 0; i<n; i++) {
            res += this;
        }
        return res;
    },

    adjustLeft: function(len, chr) { // т.е. expandToRight()
        chr = isString(chr) ? chr.charAt(0) : ' ';
        return this + chr.multiply(len - this.length);
    }
});

//
// Непосредственно начинаем реализацию машины Тьюринга
//

// Команда машины Тьюринга
var TMCommand = new Class(Object, {
    initialize: function(q,a,v,w) {
        this.q = TMCommand.parseState(q); // q  - текущее состояние
        this.a = a;                       // a  - читаемый символ
        this.v = v;                       // v  - команда или записываемый символ
        this.w = TMCommand.parseState(w); // q' - состояние, в которое переходим
    },
    // является ли команда - командой остановки машины
    isStop: function() {
        return (this.v === '#' || (this.q === this.w && this.a === this.v));
    },
    // является ли команда командой записи символа на ленту
    isWrite: function() {
        return (this.v !== '<' && this.v !== '>' && this.v !== '=' && this.v !== '#');
    },
    isLeft: function() {
        return this.v === '<';
    },
    isRight: function() {
        return this.v === '>';
    },
    elementId: function() {
        return escape('cmd_' + this.q + '_' + this.a);
    }
});
// проверяем, передали ли нам число, и если да - делаем состояние целочисленным
TMCommand.parseState = function(state) {
    if(/^\d*$/.test(state)) {
        return state.toInt(10);
    } else {
        return state.trim();
    }
};

// Лента машины Тьюринга (создаём класс для неё, хотя в любой момент времени будет лишь один его экземпляр)
var TMTape = new Class(Object, {
    initialize: function(str) {
        // обрезаем правые пробелы
        this.data = str.replace(/\s*$/, '');
        // ставим головку машины после данных
        this.pos = this.data.length;
    },
    left: function() {
        if(this.pos === 0) {
            throw('Выход за границу ленты');
        }
        this.pos -= 1;
    },
    right: function() {
        this.pos += 1;
    },
    read: function() {
        return (this.pos >= this.data.length) ? " " : this.data[this.pos];
    },
    write: function(chr) {
        if(!isString(chr) || chr.length != 1) {
            throw 'Невозможно записать на ленту "' + chr + '"';
        }
        if(this.pos >= this.data.length) {
            this.data = this.data.adjustLeft(this.pos) + chr;
        } else {
            this.data = this.data.substring(0, this.pos) + chr + this.data.substring(this.pos + 1);
        }
    },
    toHTML: function() {
        if(this.pos >= this.data.length) {
            return this.data.adjustLeft(this.pos).toHTML() + "<span class='active'>&nbsp;</span>";
        } else {
            return this.data.substring(0, this.pos).toHTML() + "<span class='active'>" + this.data[this.pos].toHTML() + "</span>" + this.data.substring(this.pos + 1).toHTML();
        }
    }
});
var tmState = 0;
var tmTape = new TMTape("");
var tmProgram = {};
var tmTimer = null;
var tmRunning = false;
var runTM = function() {
    try {
        // начальное состояние - 0
        tmState = 0;
        // заполняем данные ленты
        tmTape = new TMTape($('txt-tape').value());
        // обнуляем программу
        tmProgram = {};
        // читаем её по одной команде из поля ввода
        var prg = $('txt-program').value(), prg_html = '';
        // разбираем программу регулярным выражением
        // при разборе получим массив:
        // ['выражение целиком, нам оно не интересно', 'пробелы/переносы строк ДО команды', 'команда целиком'
        //  'начальное состояние', 'читаемый символ', 'записываемый символ/команда', 'новое состояние']
        var re = /^(\s*)(([^\s]+),(.),(.),([^\s]+))/;
        var cmd, arr;
        while(prg.trim() != "") { // пока не выцарапали все команды из текста программы
            arr = re.exec(prg);
            // если текст не пуст, а при разборе не выдано ничего, то это где-то ошибка
            // выводим первые 20 символов, чтобы пользователь мог понять, где что-то не так
            if(arr == null) {
               throw('Ошибка разбора: "' + prg.substring(0, 20) + ' ..."'); 
            }
            // создаём команду машины из элементов результата
            cmd = new TMCommand(arr[3], arr[4], arr[5], arr[6]);
            // если команда для такой ситуации уже есть - ошибка
            if(defined(tmProgram[cmd.q]) && defined(tmProgram[cmd.q][cmd.a])) {
                throw('Неоднозначная команда: состояние "' + cmd.q + '", символ "' + cmd.a + '"');
            }
            // сохраняем команду в программу
            if(!defined(tmProgram[cmd.q])) {
                tmProgram[cmd.q] = {};
            }
            tmProgram[cmd.q][cmd.a] = cmd;
            // обрезаем от программы текущую команду, чтобы перейти к разбору следующей
            prg = prg.replace(re, '');
            // к html-версии программы дописываем пробелы и переносы строк, преобразованные к html-виду
            // это нужно, чтобы программа в html-виде выглядела также как и в тексте
            prg_html += arr[1].toHTML();
            // также дописываем обёрнутую в span команду в её исходном виде (т.е. не заменяем циферные состояния на числа)
            prg_html += '<span id="' + cmd.elementId() + '">' + arr[2].toHTML() + '</span>';
        }
        $('div-program').html(prg_html);

        tmMode('run');
        
        stepTM(); // первый шаг делаем автоматически
    } catch(error) {
        setError(error);
    }
};
var stepTM = function(auto) {
    try {
        // если передали параметр auto = false или не передали auto вообще,
        // необходимо прервать выполнение машины
        if(!auto && tmTimer !== null) {
            tmTimer.stop();
            tmTimer = null;
        }
        // отрисовываем текущее положение на ленте
        // это надо сделать до шага, чтобы выделенная команда соответствовала ситуации
        $('div-tape').html(tmTape.toHTML());

        var a = tmTape.read();
        if(!defined(tmProgram[tmState])) {
            throw('Несуществующее состояние "' + tmState + '"');
        }
        if(!defined(tmProgram[tmState][a])) {
            throw('Нет команды, соответствующей состоянию "' + tmState + '" и символу "' + a + '"');
        }
        var cmd = tmProgram[tmState][a];
        // подсвечиваем текущую команду
        $(cmd.elementId()).radioClass('active');
        // выполняем команду
        if(cmd.isStop()) {
            setInfo('Машина завершила своё выполнение');
            if(tmTimer !== null) {
                tmTimer.stop();
                tmTimer = null;
            }
        } else if(cmd.isLeft()) {
            tmTape.left();
        } else if(cmd.isRight()) {
            tmTape.right();
        } else if(cmd.isWrite()) {
            tmTape.write(cmd.v);
        }
        // переходим в новое состояние
        tmState = cmd.w;
    } catch(error) {
        setError(error);
        if(tmTimer !== null) {
            tmTimer.stop();
            tmTimer = null;
        }
    }
};
var fastTM = function() {
    if(tmTimer === null) {
        tmTimer = function() { stepTM(true); }.periodical(100);
    }
};
var stopTM = function() {
    if(tmTimer !== null) {
        tmTimer.stop();
        tmTimer = null;
    }
    setError('');
    setInfo('');
    tmMode('edit');
};
var setError = function(err) {
    $('error').text(err);
};
var setInfo = function(info) {
    $('info').html(info);
};
var tmMode = function(mode) {
    if(mode == 'edit') {
        $$('.run').each('hide');
        $$('.edit').each('show');
        tmRunning = false;
    } else {
        $$('.run').each('show');
        $$('.edit').each('hide');
        tmRunning = true;
    }
};

// подготовка элементов управления при окончании загрузки html.
$(window).onReady(function() {
    tmMode('edit');

    $('txt-tape').on({
        keypress: function(event) {
            if(event.which === 13) {
                runTM();
                event.stop();
            }
        },
        focus: function() { this.select(); }
    });
    $('txt-program').on({
        keypress: function(event) {
            if(event.which === 13 && event._.ctrlKey === true) {
                runTM();
                event.stop();
            }
        },
    });
    $('div-tape').on({
        click: stopTM
    });
    $('div-program').on({
        click: stopTM
    });

    $(document).on({
        keypress: function(event) {
            if(tmRunning) {
                if(event.which === 13 && event._.ctrlKey == true) {
                    stopTM();
                    $('txt-tape').focus();
                    event.stop();
                } else if(event.which === 13) {
                    fastTM();
                    event.stop();
                } else if(event.which === 32) {
                    stepTM();
                    event.stop();
                }
            }
        }
    });

    $('btn-run').on({
        click: runTM
    });
    $('btn-stop').on({
        click: stopTM
    });
    $('btn-fast').on({
        click: fastTM
    });
    $('btn-step').on({
        click: stepTM
    });


    var toggleHelp = function() {
        $('container').toggle();
        $('div-help').toggle();
    }
    $('btn-help').on({
        click: toggleHelp
    });
    $('div-help').on({
        dblclick: toggleHelp
    });

    $('txt-program').select();
});
